/*
 *  MIDIMnp.c
 *  
 *
 *  Created by Kentaro Ueda on Thu Jun 20 2002.
 *  Copyright (c) 2002 Kentaro Ueda. All rights reserved.
 *
 */

#include <stdlib.h>
#include "SMFdefine.h"
#include "MIDIMnp.h"

extern int MIDIDataSort(long MidiDataNum, Midis *MidiData[]);
extern int CheckNoteOn(unsigned char status,unsigned char velocity);
extern int CheckNoteOff(unsigned char status,unsigned char velocity);
extern int CouplingNoteData(Midis MidiData[],long MidiDataNum,Notes *NoteData[],long *NoteDataNum);

int MIDIDataSort(long MidiDataNum, Midis *MidiData[]){

	Midis	*MidiData_sort;
	long		Number;
	int		i, j, k;
	Midis	x;

	MidiData_sort = (Midis*)malloc((sizeof(Midis))*(MidiDataNum+1));
	Number = MidiDataNum;
	
	for(i=1;i<=Number;i++)
		MidiData_sort[i] = (*MidiData)[i-1];		

	/*HeapSort*/
	for (k = Number / 2; k >= 1; k--) {
		i = k;  
		x = MidiData_sort[i];
		while ((j = 2 * i) <= Number) {
			if (j < Number && MidiData_sort[j].timestamp < MidiData_sort[j + 1].timestamp) j++;
			if (x.timestamp >= MidiData_sort[j].timestamp) break;
			MidiData_sort[i] = MidiData_sort[j]; 	
			i = j;
		}
		MidiData_sort[i] = x;
	}
	while (Number > 1) {
		x = MidiData_sort[Number];	
		MidiData_sort[Number] = MidiData_sort[1]; 
		Number--;
		i = 1;
		while ((j = 2 * i) <= Number) {
			if (j < Number && MidiData_sort[j].timestamp < MidiData_sort[j + 1].timestamp) j++;
			if (x.timestamp >= MidiData_sort[j].timestamp) break;
			MidiData_sort[i] = MidiData_sort[j]; 
			i = j;
		}
		MidiData_sort[i] = x;
	}
	
	for(i=0;i<MidiDataNum;i++)
		(*MidiData)[i] = MidiData_sort[i+1];
	
	free(MidiData_sort);
	
	return true;
}

int CheckNoteOn(unsigned char status,unsigned char velocity){

	if((status&0xf0)==0x90)
		if(velocity!=0)	return true;
	
	return false;
}

int CheckNoteOff(unsigned char status,unsigned char velocity){

	if((status&0xf0)==0x90)
		if(velocity==0)	return true;
		
	if((status&0xf0)==0x80)	return true;
	
	return false;
}

int CouplingNoteData(Midis MidiData[],long MidiDataNum,Notes *NoteData[],long *NoteDataNum){

	long MidiDataCounter;
	long s;
	long NoteCounter=0;
	unsigned char data;
	char *MidiDataFlag;

	/***initialize***/
	(*NoteDataNum)=0;
	
	/***count NoteOnData***/
	for(MidiDataCounter=0;MidiDataCounter<MidiDataNum;MidiDataCounter++)
		if(CheckNoteOn(MidiData[MidiDataCounter].status,MidiData[MidiDataCounter].velocity))
			(*NoteDataNum)+=1;

	/***malloc NoteData***/
	*NoteData = (Notes*)malloc((sizeof (Notes))*(*NoteDataNum));
	for(s=0;s<(*NoteDataNum);s++){
		(*NoteData)[s].channel		= 0;
		(*NoteData)[s].pitch 		= 0;
		(*NoteData)[s].velocity 	= 0;
		(*NoteData)[s].dummy1 	= 0;
		(*NoteData)[s].StartTime 	= 0;
		(*NoteData)[s].FinishTime	= 0;
	}
	MidiDataFlag = (char*)malloc((sizeof(char))*(MidiDataNum));
	for(s=0;s<MidiDataNum;s++)
		MidiDataFlag[s]=0;	
	/***Make MidiData pear***/
	for(MidiDataCounter=0;MidiDataCounter<MidiDataNum;MidiDataCounter++){
		if(CheckNoteOn(MidiData[MidiDataCounter].status,MidiData[MidiDataCounter].velocity)){
			(*NoteData)[NoteCounter].channel	= (unsigned char)((MidiData[MidiDataCounter].status)&0x0f);
			(*NoteData)[NoteCounter].pitch		= (unsigned char)(MidiData[MidiDataCounter].note);
			(*NoteData)[NoteCounter].velocity	= (unsigned char)(MidiData[MidiDataCounter].velocity);
			(*NoteData)[NoteCounter].StartTime	= MidiData[MidiDataCounter].timestamp;
			MidiDataFlag[MidiDataCounter]=1;
			for(s=MidiDataCounter+1;s<MidiDataNum;s++){
				if(MidiDataFlag[s]==0){
					data=MidiData[s].status;
					if((*NoteData)[NoteCounter].channel == (data&0x0f)){
						if(CheckNoteOff(MidiData[s].status,MidiData[s].velocity)){
							if((*NoteData)[NoteCounter].pitch == (unsigned char)MidiData[s].note){
								(*NoteData)[NoteCounter++].FinishTime = MidiData[s].timestamp;
								MidiDataFlag[s]=1;
								break;
							}
						}
					}
				}
			}
			if(s==MidiDataNum)
				return false;
		}
	}

	free(MidiDataFlag);
	return true;
}