/*
 *  MetaMnp.c
 *  
 *
 *  Created by Kentaro Ueda on Thu Jun 20 2002.
 *  Copyright (c) 2002 Kentaro Ueda. All rights reserved.
 *
 */

#include <stdlib.h>
#include <string.h>
#include "SMFdefine.h"
#include "MetaMnp.h"

extern int MetaDataSort(long MetaDataNum, Metas **MetaData);
extern float MetaDataToBPM(Metas MetaData);
extern int MetaDataToTimeSignature(Metas MetaData,short *denominator,short *numerator);
extern int MetaDataToKeySignatureNum(Metas MetaData);
extern int MetaDataToKeySignatureStr(Metas MetaData,char Key[4]);

extern int MetaDataSort(long MetaDataNum, Metas *MetaData[]){

	Metas	*MetaData_sort;
	long		Number;
	int		i, j, k;
	Metas	x;

	MetaData_sort = (Metas*)malloc((sizeof(Metas))*(MetaDataNum+1));
	Number = MetaDataNum;
	
	for(i=1;i<=Number;i++)
		MetaData_sort[i] = (*MetaData)[i-1];		

	/*HeapSort*/
	for (k = Number / 2; k >= 1; k--) {
		i = k;  
		x = MetaData_sort[i];
		while ((j = 2 * i) <= Number) {
			if (j < Number && MetaData_sort[j].timestamp < MetaData_sort[j + 1].timestamp) j++;
			if (x.timestamp >= MetaData_sort[j].timestamp) break;
			MetaData_sort[i] = MetaData_sort[j]; 	
			i = j;
		}
		MetaData_sort[i] = x;
	}
	while (Number > 1) {
		x = MetaData_sort[Number];	
		MetaData_sort[Number] = MetaData_sort[1]; 
		Number--;
		i = 1;
		while ((j = 2 * i) <= Number) {
			if (j < Number && MetaData_sort[j].timestamp < MetaData_sort[j + 1].timestamp) j++;
			if (x.timestamp >= MetaData_sort[j].timestamp) break;
			MetaData_sort[i] = MetaData_sort[j]; 
			i = j;
		}
		MetaData_sort[i] = x;
	}
	
	for(i=0;i<MetaDataNum;i++)
		(*MetaData)[i] = MetaData_sort[i+1];
	
	free(MetaData_sort);
	
	return true;
}

extern float MetaDataToBPM(Metas MetaData){

	long	s;
	float	BPM=0;
	
	if(MetaData.num != MetaSetTempo)	return MetaMnpErr;
	
	s=(long)((MetaData.data[0]<<16) + (0xFF00&(MetaData.data[1]<<8)) + MetaData.data[2]);
	BPM = 60000000/s;
	return BPM;
}

extern int MetaDataToTimeSignature(Metas MetaData,short *denominator,short *numerator){

	short	s;
	
	if(MetaData.num!=MetaTimeSignature)	return MetaMnpErr;
	
	*denominator=1;
	for(s=0;s<MetaData.data[0];s++)
		*denominator = (short)((*denominator)*2);
	*numerator = (short)MetaData.data[1];

	return true;
}
extern int MetaDataToKeySignatureNum(Metas MetaData){

	if(MetaData.num != MetaKeySignature)	return MetaMnpErr;

	if(MetaData.data[1]==0){
		switch(MetaData.data[0]){
			case -7 :	return BM;		break;
			case -6:	return GbM;	break;
			case -5:	return DbM;	break;
			case -4:	return AbM;	break;
			case -3:	return EbM;	break;
			case -2:	return BbM;	break;
			case -1:	return FM;		break;
			case 0:	return CM;		break;
			case 1:	return GM;		break;
			case 2:	return DM;		break;
			case 3:	return AM;		break;
			case 4:	return EM;		break;
			case 5:	return BM;		break;
			case 6:	return GbM;	break;
			case 7:	return AbM;	break;
		}
	}
	else if(MetaData.data[1]==1){
		switch(MetaData.data[0]){
			case -7:	return Abm;	break;
			case -6:	return Ebm;	break;
			case -5:	return Bbm;	break;
			case -4:	return Fm;		break;
			case -3:	return Cm;		break;
			case -2:	return Gm;		break;
			case -1:	return Dm;		break;
			case 0:	return Am;		break;
			case 1:	return Em;		break;
			case 2:	return Bm;		break;
			case 3:	return Gbm;	break;
			case 4:	return Dbm;	break;
			case 5:	return Abm;	break;
			case 6:	return Ebm;	break;
			case 7:	return Bbm;	break;
		}	
	}
	return false;
}
extern int MetaDataToKeySignatureStr(Metas MetaData,char Key[4]){

	if(MetaData.num != MetaKeySignature)	return MetaMnpErr;

	if(MetaData.data[1]==0){
		switch(MetaData.data[0]){
			case -7 :	strcpy(Key,"BM");	break;
			case -6:	strcpy(Key,"GbM");	break;
			case -5:	strcpy(Key,"DbM");	break;
			case -4:	strcpy(Key,"AbM");	break;
			case -3:	strcpy(Key,"EbM");	break;
			case -2:	strcpy(Key,"BbM");	break;
			case -1:	strcpy(Key,"FM");	break;
			case 0:	strcpy(Key,"CM");	break;
			case 1:	strcpy(Key,"GM");	break;
			case 2:	strcpy(Key,"DM");	break;
			case 3:	strcpy(Key,"AM");	break;
			case 4:	strcpy(Key,"EM");	break;
			case 5:	strcpy(Key,"BM");	break;
			case 6:	strcpy(Key,"GbM");	break;
			case 7:	strcpy(Key,"AbM");	break;
		}
	}
	else if(MetaData.data[1]==1){
		switch(MetaData.data[0]){
			case -7:	strcpy(Key,"Abm");	break;
			case -6:	strcpy(Key,"Ebm");	break;
			case -5:	strcpy(Key,"Bbm");	break;
			case -4:	strcpy(Key,"Fm");	break;
			case -3:	strcpy(Key,"Cm");	break;
			case -2:	strcpy(Key,"Gm");	break;
			case -1:	strcpy(Key,"Dm");	break;
			case 0:	strcpy(Key,"Am");	break;
			case 1:	strcpy(Key,"Em");	break;
			case 2:	strcpy(Key,"Bm");	break;
			case 3:	strcpy(Key,"Gbm");	break;
			case 4:	strcpy(Key,"Dbm");	break;
			case 5:	strcpy(Key,"Abm");	break;
			case 6:	strcpy(Key,"Ebm");	break;
			case 7:	strcpy(Key,"Bbm");	break;
		}	
	}
	return false;
}