/*
 *  TrackDataMnp.c.c
 *  
 *
 *  Created by Kentaro Ueda on Thu Jun 20 2002.
 *  Copyright (c) 2002 Kentaro Ueda. All rights reserved.
 *
 */

#include <stdlib.h>
#include <string.h>
#include "SMFdefine.h"
#include "TrackDataMnp.h"

int ReadHeaderInf(FILE *fp,short *format,short *TrackNum,short *Granulality);
int ReadTrackData(FILE *fp,char **track_data,unsigned long *track_length);
int CountDataNumber(char track_data[],unsigned long track_length,long *MidiDataNum,long *MetaDataNum);
int TrackDataOut(char track_data[],unsigned long track_length,Midis *MidiData[],Metas *MetaData[],long *MidiDataNum,long *MetaDataNum);
long	ReadVariableData(char track_data[],long *data_counter);
long	ReadTimeData(long time,char track_data[],long *data_counter);

long ReadVariableData(char track_data[],long *data_counter){
	
	long data;
	long DataCounter;
	
	DataCounter=*data_counter;
	
	data=track_data[DataCounter];
	if(track_data[DataCounter] & (char)0x80){
		data = (long)(track_data[DataCounter++]&(char)0x7f);
		do{
		data = (data << 7) + (long)(track_data[DataCounter] & (char)0x7f);	
		}
		while(track_data[DataCounter++] & (char)0x80);
	}
	else DataCounter++;
	
	*data_counter = DataCounter;
	
	return data;
}

long ReadTimeData(long time,char track_data[],long *data_counter){
	
	long data;
	long DataCounter;
	
	DataCounter=*data_counter;
	
	data = ReadVariableData(track_data,&DataCounter);
	time = time + data;
	*data_counter = DataCounter;
	
	return time;
}

int CountDataNumber(char track_data[],unsigned long track_length,long *MidiDataNum,long *MetaDataNum){
	
	long 	length=0;
	long	time=0;	
	long	data=0;
	long	data_counter;
	char	pre_status;
	
	for(data_counter=0;data_counter<track_length;){
		time = ReadTimeData(time,track_data,&data_counter);
		switch(track_data[data_counter]){
			/*case Meta_event*/
			case (char)0xFF:
			(*MetaDataNum)++;
			data_counter++;
			/*read Meta type*/
			data=(char)track_data[data_counter++];
			/*read Meta lenght*/
			length = ReadVariableData(track_data,&data_counter);
			data_counter = data_counter+length;
			break;
			
			/*case system exclusive*/
			case (char)0xF0:
			case (char)0xF7:
			(*MidiDataNum)++;
			data_counter++;
			/*read exclusive length*/
			length = ReadVariableData(track_data,&data_counter);
			data_counter=data_counter+length;	
			break;
			
			/* case normal MIDI data*/
			default:	
			(*MidiDataNum)++;
			/*if not running status*/
			if((track_data[data_counter] & (char)0x80) == (char)0x80){
				/*if program-change */
				if(((track_data[data_counter] & (char)0xF0) == (char)0xC0)
				/*or channel-pressure*/
				||((track_data[data_counter] & (char)0xF0) == (char)0xD0))
				data_counter=data_counter+2;
				else
				data_counter=data_counter+3;
				pre_status = track_data[data_counter];
			}
			/*if running status*/
			else{			
				/*if program-change */
				if(((pre_status & (char)0x00F0) == (char)0x00C0)
				/*or channel-pressure*/
				||((pre_status & (char)0x00F0) == (char)0x00D0))
				data_counter=data_counter+1;
				else
				data_counter=data_counter+2;
			}
		}
	}
		
	return true;
}

int TrackDataOut(char track_data[],unsigned long track_length,Midis *MidiData[],Metas *MetaData[],long *MidiDataNum,long *MetaDataNum){
	
	long	data_counter;
	long	time=0;	
	long	char_counter;
	long	length;
	
	for(data_counter=0;data_counter<track_length;){
		time = ReadTimeData(time,track_data,&data_counter);
		switch(track_data[data_counter]){
		/*case Meta_event*/
		case (char)0xFF:
		(*MetaData)[*MetaDataNum].timestamp = time;
		data_counter++;
		/*read Meta type*/
		(*MetaData)[*MetaDataNum].num = (char)track_data[data_counter++];
		/*read Meta length*/
		length = ReadVariableData(track_data,&data_counter);
		(*MetaData)[*MetaDataNum].length	= length;
		/*malloc meta data*/
		((*MetaData)[*MetaDataNum].data) = (char*)malloc((sizeof (char))*(length+2));
		if(((*MetaData)[*MetaDataNum].data)==NULL)
			return MemoryErr;
		/*reaed meta data*/
		for(char_counter=0;char_counter<length;char_counter++)
			((*MetaData)[*MetaDataNum].data)[char_counter]
				= (char)track_data[data_counter++]; 			
		(*MetaDataNum)++;
		break;
		
		/*case system exclusive*/    
		case (char)0xF0:
		case (char)0xF7:
			(*MidiData)[*MidiDataNum].timestamp = time;
			/*read status*/
			if(track_data[data_counter] == (char)0xF0)
				(*MidiData)[*MidiDataNum].status = (unsigned char)0xF0;
			if(track_data[data_counter] == (char)0xF7)
				(*MidiData)[*MidiDataNum].status = (unsigned char)0xF7;
			data_counter++;	
			/*read system exclusive length*/
			length = ReadVariableData(track_data,&data_counter);
			(*MidiData)[*MidiDataNum].exclusiveLength = length;
			/*malloc system exclusive data*/ 
			((*MidiData)[*MidiDataNum].exclusiveData)
				= (char*)malloc((sizeof (char))*(length+2));
			if(((*MidiData)[*MidiDataNum].exclusiveData)==NULL)
				return MemoryErr;
			(*MidiData)[*MidiDataNum].note = 0;
			(*MidiData)[*MidiDataNum].velocity = 0;
			/*read system exclusive data*/
			for(char_counter=0;char_counter<length;char_counter++)
				((*MidiData)[*MidiDataNum].exclusiveData)[char_counter]
					= (char)track_data[data_counter++];
			(*MidiDataNum)++;
			break;
			
		/*case normal MIDI data*/    
		default: 
			(*MidiData)[*MidiDataNum].timestamp = time;
			(*MidiData)[*MidiDataNum].exclusiveLength = 0;
			(*MidiData)[*MidiDataNum].exclusiveData = NULL;
			/*if not running status*/
			if((track_data[data_counter] & (char)0x80) == (char)0x80){
				(*MidiData)[*MidiDataNum].status
					= (unsigned char)track_data[data_counter++];
				(*MidiData)[*MidiDataNum].note
					= (unsigned char)track_data[data_counter++];
				/*if program-change */
				if(((track_data[data_counter-2] & (char)0xF0) == (char)0xC0)
				/*or channel-pressure*/
				||((track_data[data_counter-2] & (char)0xF0) == (char)0xD0))
				(*MidiData)[*MidiDataNum].velocity= 0;
				else
				(*MidiData)[*MidiDataNum].velocity 
					= (unsigned char)track_data[data_counter++];						
		}
		/*if running status*/
		else{
			(*MidiData)[*MidiDataNum].status
				= (unsigned char)((*MidiData)[(*MidiDataNum)-1].status);
			(*MidiData)[*MidiDataNum].note
				= (unsigned char)track_data[data_counter++];
			/*if program-change */
			if(((track_data[data_counter-2] & (char)0xF0) == (char)0xC0)
			/*or channel-pressure*/
			||((track_data[data_counter-2] & (char)0xF0) == (char)0xD0))
			(*MidiData)[*MidiDataNum].velocity= 0;
			else
			(*MidiData)[*MidiDataNum].velocity= (unsigned char)track_data[data_counter++];						
		}
		(*MidiDataNum)++;
		break;
		}
	}
	return true;
}

/******** For SMFOut ***********************************************/
int WriteHeaderInf(FILE *fp,int format,short TrackNum,short Granulality);
int MMDataSort(long MMDataNum, MMDatas *MMData[]);
int CalcVariableData(long data);
int MixMIDIandMetaData(Midis MidiData[],Metas MetaData[],long MidiDataNum,long MetaDataNum,MMDatas *MMData[],unsigned long *TrackLength);
int MakeTrackData(char *TrackData[],Midis MidiData[],Metas MetaData[],long MidiDataNum,long MetaDataNum,MMDatas MMData[],unsigned long TrackLength);
int ChangeVariableData(long data,char DataString[],short *length);

int MMDataSort(long MMDataNum, MMDatas *MMData[]){
	
	/***use heap sort***/
	MMDatas	*MMData_sort;
	long		Number;
	int		i, j, k;
	MMDatas	x;
	
	MMData_sort = (MMDatas*)malloc((sizeof(MMDatas))*(MMDataNum+1));
	Number = MMDataNum;
	
	for(i=1;i<=Number;i++)
		MMData_sort[i] = (*MMData)[i-1];
	/*HeapSort*/
	for (k = Number / 2; k >= 1; k--) {
		i = k;  
		x = MMData_sort[i];
		while ((j = 2 * i) <= Number) {
		if (j < Number && MMData_sort[j].timestamp < MMData_sort[j + 1].timestamp) j++;
		if (x.timestamp >= MMData_sort[j].timestamp) break;
			MMData_sort[i] = MMData_sort[j]; 	
			i = j;
		}
		MMData_sort[i] = x;
	}
	while (Number > 1) {
		x = MMData_sort[Number];	
		MMData_sort[Number] = MMData_sort[1]; 
		Number--;
		i = 1;
		while ((j = 2 * i) <= Number) {
		if (j < Number && MMData_sort[j].timestamp < MMData_sort[j + 1].timestamp) j++;
		if (x.timestamp >= MMData_sort[j].timestamp) break;
		MMData_sort[i] = MMData_sort[j]; 
		i = j;
		}
		MMData_sort[i] = x;
	}
	
	for(i=0;i<MMDataNum;i++)
		(*MMData)[i] = MMData_sort[i+1];
	
	free(MMData_sort);
	
	return true;
}

int CalcVariableData(long data){

	if(data & 0x0FE0000)	return 4;
	if(data & 0x001FC000)	return 3;
	if(data & 0x00004F80)	return 2;
	else			return 1;
}

int MixMIDIandMetaData(Midis MidiData[],Metas MetaData[],long MidiDataNum,long MetaDataNum,MMDatas *MMData[],unsigned long *TrackLength){
		
	long	num;
	short	preStatus;
	
	(*TrackLength) = 0;
	/*Input MIDI&MetaData to MMDatas*/
	(*MMData) = (MMDatas*)malloc((sizeof(MMDatas))*(MidiDataNum+MetaDataNum));
	if((*MMData)==NULL)
		return MemoryErr;
		
	for(num=0;num<MidiDataNum;num++){
		(*MMData)[num].datalength = 0;
		(*MMData)[num].status = (short)MidiData[num].status;
		(*MMData)[num].type = typeMIDI;
		(*MMData)[num].num = num;
		(*MMData)[num].timestamp = MidiData[num].timestamp;
	}
	for(num=MidiDataNum;num<(MidiDataNum+MetaDataNum);num++){
		(*MMData)[num].datalength
			= 2+CalcVariableData(MetaData[num-MidiDataNum].length)
				+MetaData[num-MidiDataNum].length;
		(*MMData)[num].status = 0;
		(*MMData)[num].type = typeMeta;
		(*MMData)[num].num = num-MidiDataNum;
		(*MMData)[num].timestamp = MetaData[num-MidiDataNum].timestamp;
	}
	/*Sort MMDatas*/
	MMDataSort((MidiDataNum+MetaDataNum),MMData);
	/*Search Running Status*/
	for(num=(MidiDataNum+MetaDataNum);num>0;num--)
		if(((*MMData)[num].type == (*MMData)[num-1].type)&&((*MMData)[num-1].type== typeMIDI))
		if((*MMData)[num].status == (*MMData)[num-1].status)
			(*MMData)[num].status = RunningStatus;
	/*Calc MIDI-data length*/
	for(num=0;num<(MidiDataNum+MetaDataNum-1);num++){
		if((*MMData)[num].type == typeMIDI){
		switch((*MMData)[num].status){
		/*case system exclusive*/               
		case 0xF0:
		case 0xF7:
			(*MMData)[num].datalength
			= 1+CalcVariableData(MidiData[(*MMData)[num].num].exclusiveLength)
				+MidiData[(*MMData)[num].num].exclusiveLength;
			break;
		/*case running status*/
		case RunningStatus:
			/*if program-change*/
			if(((preStatus & (short)0xF0) == (short)0xC0)
			/*or channel-pressure*/
			||((preStatus & (short)0xF0) == (short)0xD0))
			(*MMData)[num].datalength = 1;
			else 
			(*MMData)[num].datalength = 2;
			break;
		/*normal MIDI data*/
		default :
			/*if program-change*/
			if((((*MMData)[num].status & (short)0xF0) == (short)0xC0)
			/*or channel-pressure*/
			||(((*MMData)[num].status & (short)0xF0) == (short)0xD0))
			(*MMData)[num].datalength = 2;
			else 
			(*MMData)[num].datalength = 3;
			preStatus = (*MMData)[num].status;
			break;			
		}
		}
	}
	/*Calc delta-time and its datalength*/
	for(num=(MidiDataNum+MetaDataNum-1);num>0;num--)
		(*MMData)[num].timestamp
			= (*MMData)[num].timestamp-(*MMData)[num-1].timestamp;
	for(num=0;num<(MidiDataNum+MetaDataNum);num++)
		(*MMData)[num].datalength
			= (*MMData)[num].datalength + CalcVariableData((*MMData)[num].timestamp);
	/*Calc TrackLength*/
	for(num=0;num<(MidiDataNum+MetaDataNum);num++)
		(*TrackLength) = (*TrackLength) + (*MMData)[num].datalength;
	return true;
}


int ChangeVariableData(long data,char DataString[],short *length){

	if(data & 0x0FE0000){
		(*length) = 4;
		DataString[0] = (char)((char)(data>>21)+(char)0x80);
		DataString[1] = (char)((char)(data>>14)+(char)0x80);
		DataString[2] = (char)((char)(data>>7)+(char)0x80);
		DataString[3] = (char)(data&0x7F);
	}
	else if(data & 0x001FC000){
		(*length) = 3;
		DataString[0] = (char)((char)(data>>14)+(char)(0x80));
		DataString[1] = (char)((char)(data>>7)+(char)(0x80));
		DataString[2] = (char)(data&0x7F);
	}
	else if(data & 0x00004F80){
		(*length) = 2;
		DataString[0] = (char)((char)(data>>7)+(char)(0x80));
		DataString[1] = (char)((char)(data&0x7F));
	}
	else{
		(*length) = 1;
		DataString[0] = (char)(data);
	}
	return true;
}


int MakeTrackData(char *TrackData[],Midis MidiData[],Metas MetaData[],long MidiDataNum,long MetaDataNum,MMDatas MMData[],unsigned long TrackLength){
	
	long	num,n;
	long	data_counter=0;
	char	DataString[4];
	short	length;
	
	/*Malloc TrackData+"MTrk"+TrackLength*/
	(*TrackData) = (char*)malloc((sizeof(char))*(TrackLength+4+4));
	if((*TrackData == NULL))
		return MemoryErr;
		
	/*write "MTrk"*/
	strcpy((*TrackData),"MTrk");							
	data_counter = 4;
	/*write TrackLength*/ 
	(*TrackData)[data_counter++] = (char)(TrackLength>>24);
	(*TrackData)[data_counter++] = (char)(TrackLength>>16);
	(*TrackData)[data_counter++] = (char)(TrackLength>>8);
	(*TrackData)[data_counter++] = (char)(TrackLength);
	
	for(num=0;num<(MidiDataNum+MetaDataNum);num++){
		/*if Meta Data*/
		if(MMData[num].type == typeMeta){
			/*write timestamp data*/
			ChangeVariableData(MMData[num].timestamp,DataString,&length);
			for(n=0;n<length;n++)
				(*TrackData)[data_counter++] = DataString[n];
			/*write meta-event*/
			(*TrackData)[data_counter++] = (char)0xFF;
			/*write meta-num(type)*/
			(*TrackData)[data_counter++] = (char)MetaData[MMData[num].num].num;
			/*write meta-length*/
			ChangeVariableData(MetaData[MMData[num].num].length,DataString,&length);
			for(n=0;n<length;n++)
				(*TrackData)[data_counter++] = DataString[n];
			/*write meta-data*/
			for(n=0;n<MetaData[MMData[num].num].length;n++)
				(*TrackData)[data_counter++] = MetaData[MMData[num].num].data[n];	
		}
		/*if MIDI Data*/
		else if(MMData[num].type == typeMIDI){
			/*write timestamp data*/
			ChangeVariableData(MMData[num].timestamp,DataString,&length);
			for(n=0;n<length;n++)
				(*TrackData)[data_counter++] = DataString[n];
				
			/*write midi-event if not running status*/
			if(MMData[num].status != -1)
				(*TrackData)[data_counter++] = (char)MMData[num].status;
			/*if system-exclusive*/
			if(((char)MMData[num].status==(char)0xF0)
				||((char)MMData[num].status==(char)0xF7)){
				/*write exclusive length*/
				ChangeVariableData(MidiData[MMData[num].num].exclusiveLength,
					DataString,&length);
				for(n=0;n<length;n++)
				(*TrackData)[data_counter++] = DataString[n];
				/*write exclusive data*/
				for(n=0;n<MidiData[MMData[num].num].exclusiveLength;n++)
				(*TrackData)[data_counter++]
					= MidiData[MMData[num].num].exclusiveData[n];			           	}
			/*if normal MIDI-Data*/
			else{
				/*write note*/
				(*TrackData)[data_counter++] = (char)(MidiData[MMData[num].num].note);
				/*write velocity if not program-change or controle-change*/
				/*if not program-change */
				if((((char)MMData[num].status & (char)0xF0) != (char)0xC0)
				/*and not channel-pressure*/
				&&(((char)MMData[num].status & (char)0xF0) != (char)0xD0))
				(*TrackData)[data_counter++]
					= (char)(MidiData[MMData[num].num].velocity);
			}
		}
	}	
	if(data_counter != TrackLength + 4 + 4)	return false;
	
	return true;
}
